"""
Unit tests for BigipTask class
"""

from typing import Any, Dict, List, Optional

import pytest

from certbot_deployer_bigip.certbot_deployer_bigip import BigipTask


@pytest.mark.parametrize(
    "args, passed_args",
    [
        ({"arg1": "arg1_value", "arg2": "arg2_value"}, ["arg1_value", "arg2_value"]),
        ({}, [None, None]),
    ],
)
def test_bigiptask_exec(args: Dict[str, str], passed_args: List[str]) -> None:
    """
    Verify that a task calls `exec_function` with `exec_args`
    """
    called: bool = False
    called_args: list = []

    def some_callable(
        *, arg1: Optional[Any] = None, arg2: Optional[Any] = None
    ) -> None:
        # pylint: disable-next=unused-variable
        nonlocal called
        called = True
        called_args.extend((arg1, arg2))

    # pylint: disable-next=unused-variable
    task = BigipTask(
        exec_function=some_callable,
        exec_kwargs=args,
    )
    task.execute()
    assert called is not False
    assert called_args == passed_args



"""
Unit tests for the `BigipCertificateBundle` class
"""

from pathlib import Path

import pytest

from certbot_deployer import CERT_FILENAME
from certbot_deployer.test_helpers import COMMON_NAME, NOT_VALID_AFTER
from certbot_deployer_bigip.certbot_deployer_bigip import BigipCertificateBundle

# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=too-few-public-methods
# pylint: disable=attribute-defined-outside-init

STATIC_TEST_CERT: str = """-----BEGIN CERTIFICATE-----
MIIC2TCCAcGgAwIBAgIUD5bm1RAbxJ7dqTIlZL2GwF+B8FkwDQYJKoZIhvcNAQEL
BQAwGzEZMBcGA1UEAwwQdGVzdCBjb21tb24gbmFtZTAgFw0yMDAxMDEwMDAwMDBa
GA8yMDk5MDEwMTAwMDAwMFowGzEZMBcGA1UEAwwQdGVzdCBjb21tb24gbmFtZTCC
ASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAMdnRQoxdW9vrQcXlH1MsDuA
Vi/mc/Tq6LJB6wKoaHLiT+aVa0Md1IQtlceL/Y9OMdjWiRR5xex2lsRZ0IdcMO+7
1txrG2u2NmZ6HSG8CyzMZmPbUClLYySwv7aLgj76Q3I3E/WVnatja9W1PQxSw0Sz
dDsC2L5jc4hiD/tCAR0PvGv8eiAnjYhIsKfU5Odcoe5dw4YGqmYRbcBYqRAisQgy
ekB1VXqsRlmqm+HQSwwon7rL0p4Nzaub7ZTM7qsc8dWs3RFZlWzFRrfrwo2Le7zh
gPZdul5rz9JUcBttOVqZVES8xKG3+6iLU6/RlGsWFyyaflxWjtU0GAJZWW/ho28C
AwEAAaMTMBEwDwYDVR0TAQH/BAUwAwEB/zANBgkqhkiG9w0BAQsFAAOCAQEAk5N/
c6nfPbCMzvs1SGzkTyN6gyY7dnZwuIPSbr0iN1+tWC3UJclGRPCMqD62/T+UjKYA
6VNyTkUqMmib0qbHo67kfidEGe+A1BV9v0MxdVaa8WAop7D6PtHMLNHiIQBYkc02
QvcY6OT8evsG6zbiR8+Tj4pMIBom0AHTt3HqSfoynEpyonwutVJxwQpAqs6ZxAGT
mgNgBLNbTLrwk5ScmaASmjl3Ua24QRT48oK16Gz/JjrRv1TgEl7NBX0fPuljjyhw
qSlAqJRr0ANRfarf+M62Cq9qV/JzKFDGlPvAY7aQemMaIXrqwEPuLRP3SvBL+NBm
ahH6YxA25wiuaynPpw==
-----END CERTIFICATE-----
"""
STATIC_TEST_CERT_FINGERPRINT: str = (
    # pylint: disable-next=line-too-long
    "SHA256/44:D6:7E:66:11:B4:16:8C:5A:FD:36:F1:24:8E:18:30:60:5F:98:47:7C:1C:B3:1B:9F:76:74:8F:A5:62:9D:92"
)


@pytest.fixture(name="setup_bundle_path", scope="function")
def fixture_setup_bundle_path(tmp_path: Path) -> Path:
    """
    Given a pytest temp folder, reate a "valid" certificate there for the
    BigipCertificateBundle class to operate upon
    """
    with open(tmp_path / CERT_FILENAME, "w", encoding="utf-8") as cert_file:
        cert_file.write(STATIC_TEST_CERT)
    return tmp_path


def test_bundle_fingerprint(setup_bundle_path: Path) -> None:
    """
    Test that the cert bundle creates its fingerprint as the f5 expects

    we'll use a hard-coded cert and fingerprint instead of jut re-implementing
    the fingerprint code from the module
    """
    bundle: BigipCertificateBundle = BigipCertificateBundle(path=str(setup_bundle_path))
    assert bundle.fingerprint == STATIC_TEST_CERT_FINGERPRINT


def test_bundle_autoname(setup_bundle_path: Path) -> None:
    """test"""
    bundle: BigipCertificateBundle = BigipCertificateBundle(path=str(setup_bundle_path))
    assert bundle.name == f"{COMMON_NAME}.{NOT_VALID_AFTER.isoformat()}"



"""
Package metadata
"""

__author__ = "IAS"
__version__ = "0.5.1"
__title__ = "certbot_deployer_bigip"
__license__ = "License :: OSI Approved :: MIT License"
__description__ = (
    "Certbot Deployer plugin for deploying certificates to F5 BIG-IP devices"
)



"""
BIG-IP specific deployer code
"""

import argparse
import logging
import os
import posixpath
import re
import sys
import subprocess
import tempfile
import textwrap

# import sys
import warnings


from collections import namedtuple
from typing import Any, Callable, ClassVar, Dict, Iterable, List, Optional

import paramiko  # type: ignore
import scp as paramiko_scp  # type: ignore

# Suppress warnings thanks to old Python and fabric
# pylint: disable-next=duplicate-code
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=UserWarning)
    from cryptography.utils import CryptographyDeprecationWarning
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)
    from cryptography import x509
    from cryptography.hazmat.primitives import hashes
    from fabric import Config, Connection  # type:ignore
from invoke.exceptions import UnexpectedExit

from certbot_deployer import main as framework_main
from certbot_deployer import Deployer, CertificateBundle, CertificateComponent
from certbot_deployer_bigip.meta import __description__

BIGIP_FINGERPRINT_ALGO: str = "SHA256"
BIGIP_FINGERPRINT_ALGO_FUNC: Callable = hashes.SHA256


CertProfile = namedtuple("CertProfile", "name type")


# pylint: disable-next=too-few-public-methods
class BigipCertificateBundle(CertificateBundle):
    """
    CertificateBundle but with the specific fingerprint format used by the F5
    """

    def __init__(self, *args: Any, name: Optional[str] = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        safe_name: str = re.sub(r"[^.a-zA-Z0-9_-]", "_", self.common_name)
        self.name = f"{safe_name}.{self.expires}" if name is None else name
        self.fingerprint: str = self._fingerprint(self.certdata)

    @staticmethod
    def _fingerprint(certdata: x509.Certificate) -> str:
        """
        Take an x509 cert obj and return the fingerprint in the format used by
        the BIG-IP
        """
        fingerprint: str = (
            certdata.fingerprint(BIGIP_FINGERPRINT_ALGO_FUNC()).hex().upper()
        )
        fingerprint = f"{BIGIP_FINGERPRINT_ALGO}/" + ":".join(
            fingerprint[i : i + 2] for i in range(0, len(fingerprint), 2)
        )
        return fingerprint


class BigipTask:
    """
    @ single BIG-IP task for the deployer to run
    """

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        exec_function: Callable,
        name: Optional[str] = None,
        exec_args: Optional[Iterable[Any]] = None,
        exec_kwargs: Optional[Dict[str, Any]] = None,
        revert_function: Optional[Callable] = None,
        revert_args: Optional[Iterable[Any]] = None,
        revert_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """init"""
        self.name: Optional[str] = name
        self.exec_function: Callable = exec_function
        self.exec_kwargs: Dict[str, Any] = (
            exec_kwargs if exec_kwargs is not None else {}
        )
        self.exec_args: Iterable[Any] = exec_args if exec_args is not None else []
        self.revert_function: Optional[Callable] = revert_function
        self.revert_kwargs: Iterable[Any] = (
            revert_kwargs if revert_kwargs is not None else []
        )
        self.revert_args: Iterable[Any] = revert_args if revert_args is not None else []

    def __eq__(self, othertask: object) -> bool:
        if not isinstance(othertask, BigipTask):
            return NotImplemented
        res: bool
        try:
            res = (
                self.exec_args,
                self.exec_function,
                self.exec_kwargs,
                self.revert_args,
                self.revert_function,
                self.revert_kwargs,
            ) == (
                othertask.exec_args,  # type:ignore
                othertask.exec_function,  # type:ignore
                othertask.exec_kwargs,  # type:ignore
                othertask.revert_args,  # type:ignore
                othertask.revert_function,  # type:ignore
                othertask.revert_kwargs,  # type:ignore
            )
        except AttributeError:
            res = False
        return res

    def execute(self, *args: Any, **kwargs: Any) -> None:
        """
        Entrypoint for this task to actually run
        """
        modargs: list = list(args)
        modargs.extend(self.exec_args)
        modkwargs = dict(kwargs)
        modkwargs.update(self.exec_kwargs)
        self.exec_function(*modargs, **modkwargs)

    def revert(self, *args: Any, **kwargs: Any) -> None:
        """
        Revert the corresponding changes if possible
        """
        if self.revert_function is not None:
            self.revert_function(*args, **kwargs)


# pylint: disable-next=too-many-instance-attributes
class BigipDeployer(Deployer):
    """
    BigIP cert deployer
    """

    subcommand: ClassVar[str] = "bigip"

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        *,
        host: str,
        dest_temp_dir: str,
        certificate_bundle: BigipCertificateBundle,
        sync_group: Optional[str],
        profile: Optional[CertProfile],
    ) -> None:
        self.host: str = host
        self.dest_temp_dir: str = dest_temp_dir
        self.sync_group: Optional[str] = sync_group
        self.certificate_bundle: BigipCertificateBundle = certificate_bundle
        self.profile: Optional[CertProfile] = profile
        logging.debug("BigIp config initialized as: `%s`", str(self))
        self.operations: list = []

        self.conn: Connection = Connection(
            host=self.host,
            config=Config(
                overrides={
                    "run": {"hide": True},
                }
            ),
        )

    @staticmethod
    def register_args(*, parser: argparse.ArgumentParser) -> None:
        """
        Modify argparse subparser for this deployer target
        """
        parser.formatter_class = argparse.RawDescriptionHelpFormatter
        parser.description = f"""BIG-IP subcommand
        {__description__}
        """
        parser.epilog = textwrap.dedent(
            """
        Examples:

        All examples assume the tool is being run as a Certbot deploy hook, and
        the environment variable `RENEWED_LINEAGE` points to the live
        certificate directory just updated by Certbot.

            # Install certificate on the BIG-IP device named with its expiry timestamp
            # as `host.domain.tld.YYY-MM-DDTHH:MM:SS` (the default)

            %(prog)s --host bigip.domain.tld

            # Install certificate on the BIG-IP device and then synchronize
            # device changes to a device group

            %(prog)s --host bigip.domain.tld --sync-group yoursyncgroup

            # Install certificate on the BIG-IP device and then associate it
            # with a client-ssl profile. If the profile does not exist, it will be created.

            %(prog)s --host bigip.domain.tld --profile-name yourprofile \\
                --profile-type client-ssl

            # Print out the deployment tasks that would be taken, but do not run them

            %(prog)s --host bigip.domain.tld --dry-run

        """
        )

        parser.add_argument(
            "--host",
            "-H",
            default=os.environ.get("HOST", None),
            help=("BIG-IP host to target with changes."),
            type=str,
        )

        parser.add_argument(
            "--dest-temp-dir",
            "-t",
            default=os.environ.get("DEST_TEMP_DIR", "/var/tmp"),
            help=(
                "The temp path on the BIG-IP to use when uploading the certificates for "
                "installation. This tool will try to zero out the certificates at the end "
                "of the run once they are ingested into the BIG-IP config, but you should "
                'probably try to put them in a "safe" place. Default: `/var/tmp`'
            ),
            type=str,
        )

        parser.add_argument(
            "--cert-name",
            "-c",
            default=None,
            help=(
                "The name to give the certificate objects on the BIG-IP. If not given, "
                "this tool will use the Common Name from the certificate with the "
                "ISO8601-formatted `Not After` value appended, e.g. "
                "`host.domain.tld.YYY-MM-DDTHH:MM:SS`"
            ),
            type=str,
        )

        parser.add_argument(
            "--profile-name",
            "-p",
            help=(
                "The name of the profile to modify/create to use the deployed certificate"
                "If given, the corresponding `--profile-type` is also required. If "
                "neither are given, this tool will not create/modify any profiles."
            ),
            type=str,
        )

        parser.add_argument(
            "--profile-type",
            "-r",
            help=(
                "The type of profile to create if necessary e.g. `client-ssl`. See "
                "BIG-IP docs for details on all profile types. If given, the "
                "corresponding `--profile-name` is also required. If neither are given, "
                "this tool will not create/modify any profiles."
            ),
            type=str,
        )

        parser.add_argument(
            "--sync-group",
            "-s",
            help=(
                "The sync-group to synchronize to after the certificates are deployed. "
                "This tool will not try to sync the BIG-IP node if this argument is not "
                "given"
            ),
            type=str,
        )

        parser.add_argument(
            "--dry-run",
            "-d",
            action="store_true",
            default=False,
            help=(
                "Report the workflow steps that would run without actually running them"
            ),
        )

    def __str__(self) -> str:
        """represent self"""
        return str(vars(self))

    @staticmethod
    def argparse_post(*, args: argparse.Namespace) -> None:
        """
        Checks for after arguments are otherwise fully processed
        """
        args_vars: Dict[str, Any] = vars(args)
        if [
            args_vars.get("profile_name"),
            args_vars.get("profile_type"),
        ].count(None) == 1:
            raise argparse.ArgumentTypeError(
                "If either `--profile` or `--profile-type` are passed, both are required."
            )
        if not args_vars.get("host"):
            raise argparse.ArgumentTypeError("`--host/H is required")

    def get_workflow(self) -> List[BigipTask]:
        """
        Return workflow for this deployer
        """
        tasks: List[BigipTask] = []
        if self.sync_group is not None:
            tasks.append(BigipTask(name="Verify Sync", exec_function=self.verify_sync))
        tasks.extend(
            [
                BigipTask(
                    name="Put cert file to remote",
                    exec_function=self.put_bigip_file,
                    exec_kwargs={"component": self.certificate_bundle.fullchain},
                ),
                BigipTask(
                    name="Install cert",
                    exec_function=self.install_cert,
                    exec_kwargs={"component": self.certificate_bundle.fullchain},
                ),
                BigipTask(
                    name="Verify cert installed",
                    exec_function=self.verify_cert_installed,
                    exec_kwargs={"component": self.certificate_bundle.fullchain},
                ),
                BigipTask(
                    name="Put key file to remote",
                    exec_function=self.put_bigip_file,
                    exec_kwargs={"component": self.certificate_bundle.key},
                ),
                BigipTask(
                    name="Install key",
                    exec_function=self.install_cert,
                    exec_kwargs={"component": self.certificate_bundle.key},
                ),
                BigipTask(
                    name="Verify key installed",
                    exec_function=self.verify_cert_installed,
                    exec_kwargs={"component": self.certificate_bundle.key},
                ),
                BigipTask(
                    name="Zero out cert file on remote",
                    exec_function=self.zero_bigip_file,
                    exec_kwargs={"component": self.certificate_bundle.fullchain},
                ),
                BigipTask(
                    name="Zero out key file on remote",
                    exec_function=self.zero_bigip_file,
                    exec_kwargs={"component": self.certificate_bundle.key},
                ),
                BigipTask(
                    name="Save running config to disk",
                    exec_function=self.save,
                ),
            ]
        )
        if self.profile is not None:
            tasks.append(
                BigipTask(
                    name="Create/modify profile",
                    exec_function=self.manage_profile,
                    exec_kwargs={"component": self.certificate_bundle.fullchain},
                )
            )
        if self.sync_group is not None:
            tasks.append(
                BigipTask(
                    name="Sync",
                    exec_function=self.sync,
                ),
            )
        return tasks

    # pylint: disable-next=unused-argument
    def verify_sync(self, *args: Any, **kwargs: Any) -> None:
        """
        Ensure in sync
        """
        try:
            # Check node sync status
            logging.info("Checking sync status on remote...")
            cmd: str = "show /cm sync-status"
            logging.debug("`%s`", cmd)
            res = self.conn.run(cmd)
            if not re.search("^Status.*In Sync$", res.stdout, re.I | re.MULTILINE):
                raise RuntimeError(
                    (
                        "BIG-IP configuration out of sync. Cannot continue. All "
                        "further operations aborted."
                    )
                )
        except UnexpectedExit as err:
            raise RuntimeError(
                "Failed checking sync status. Aborting all further operations."
            ) from err

    def _scp(self, localpath: str, remotepath: str) -> None:
        """
        Send a file via SCP.

        In general these days, SCP tools actually use the SFTP protocol. This
        includes the Fabric module on which we rely for running remote
        commands.

        For our target F5 BIG-IP devices, the opposite is true - *only* the SCP
        protocol (such as it is) is supported for `tmsh` users:

        https://my.f5.com/manage/s/article/K22885182

        So this method utilizes the separate `scp` module to implement the SCP
        protocol which requires a tiny bit of extra handling.
        """
        if not self.conn.transport:
            # The connection "transport," which we need for our scp client, may
            # not exist yet if no commands have yet run over than connection
            self.conn.open()
        # Avoiding `with ..as` here because though it would be more elegant, it
        # would make testing a bit more complex, requiring it to implement a
        # context manager and meh let's not
        scp_client = paramiko_scp.SCPClient(self.conn.transport)
        scp_client.put(localpath, remotepath)
        scp_client.close()

    def put_bigip_file(
        # pylint: disable-next=unused-argument
        self,
        component: CertificateComponent,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Put the cert or key on the remote
        """
        logging.info(
            "Putting %s `%s` over scp to `%s`",
            component.label,
            component.path,
            self.dest_temp_dir,
        )
        self._scp(
            localpath=component.path,
            remotepath=posixpath.join(self.dest_temp_dir, component.filename),
        )

    def install_cert(
        # pylint: disable-next=unused-argument
        self,
        component: CertificateComponent,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Install cert

        Note that any cert, full-chain or leaf, is still given the type `cert`
        in the TMSH install command
        """
        remote_filepath: str = posixpath.join(self.dest_temp_dir, component.filename)
        bigip_cert_type = "cert" if component.label in ["cert", "fullchain"] else "key"
        try:
            # Install the cert
            logging.info("Installing %s", bigip_cert_type)
            cmd: str = (
                f"install /sys crypto {bigip_cert_type} {self.certificate_bundle.name} "
                f"from-local-file {remote_filepath}"
            )
            logging.debug("`%s`", cmd)
            self.conn.run(cmd)
        except UnexpectedExit as err:
            raise RuntimeError(
                f"Failed installing {component.label} `{self.certificate_bundle.name}` "
                "on remote"
            ) from err

    def verify_cert_installed(
        # pylint: disable-next=unused-argument
        self,
        component: CertificateComponent,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Verify cert
        """
        bigip_cert_type = "cert" if component.label in ["cert", "fullchain"] else "key"
        try:
            # Verify the cert installed
            cmd: str = (
                f"list /sys crypto {bigip_cert_type} {self.certificate_bundle.name}"
            )
            logging.debug("`%s`", cmd)
            res = self.conn.run(cmd)
            if (
                bigip_cert_type == "cert"
                and not self.certificate_bundle.fingerprint in res.stdout
            ):
                # Unfortunately we don't really have a way to verify the state
                # of the key via `tmsh` other than that it is present by name,
                # but at least we can verify the cert
                raise RuntimeError(
                    "Failed to match certificate fingerprint in "
                    f"`{self.certificate_bundle.name}` on remote"
                )
        except UnexpectedExit as err:
            raise RuntimeError(
                f"Failed verifying {bigip_cert_type} `{self.certificate_bundle.name}` "
                "present on remote"
            ) from err

    def zero_bigip_file(
        # pylint: disable-next=unused-argument
        self,
        component: CertificateComponent,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        "Wipe" file on remote by zeroing it out, for inability to just remove the
        file without admin
        """
        remote_filepath: str = posixpath.join(self.dest_temp_dir, component.filename)
        with tempfile.NamedTemporaryFile(mode="r") as emptyfile:
            logging.info(
                "Putting empty file `%s` to remote `%s` over scp...",
                emptyfile.name,
                remote_filepath,
            )
            self._scp(
                localpath=emptyfile.name,
                remotepath=remote_filepath,
            )

    # pylint: disable-next=unused-argument
    def manage_profile(self, *args: Any, **kwargs: Any) -> None:
        """
        Modify or, failing that, create profile to include the certs
        """
        cmd: str
        if self.profile is None:
            raise RuntimeError(
                "This method cannot modify or create a remote profile unless its details "
                "are provided"
            )
        try:
            logging.info("Checking for existing profile on BIG-IP..")
            # This tmsh command exits nonzero if the object is not found
            cmd = f"list /ltm profile {self.profile.type} {self.profile.name}"
            logging.debug("`%s`", cmd)
            self.conn.run(cmd)
        except UnexpectedExit:
            try:
                cmd = (
                    f"create /ltm profile {self.profile.type} {self.profile.name} "
                    f"cert {self.certificate_bundle.name} key {self.certificate_bundle.name}"
                )
                logging.debug("`%s`", cmd)
                self.conn.run(cmd)
            except UnexpectedExit as ierr:
                raise RuntimeError(
                    f"Failed to create profile `{self.profile.name}` on remote"
                ) from ierr
        else:
            try:
                cmd = (
                    f"modify /ltm profile {self.profile.type} {self.profile.name} cert "
                    f"{self.certificate_bundle.name} key {self.certificate_bundle.name}"
                )
                logging.debug("`%s`", cmd)
                self.conn.run(cmd)
            except UnexpectedExit as err:
                raise RuntimeError(
                    "Unexpected failure when trying to update the profile "
                    f"`{self.profile.name}` to use the new cert"
                ) from err

    # pylint: disable-next=unused-argument
    def save(self, *args: Any, **kwargs: Any) -> None:
        """
        Save new running config to disk (`bigip.conf`)
        """
        try:
            cmd: str = "save /sys config"
            logging.debug("`%s`", cmd)
            self.conn.run(cmd)
        except UnexpectedExit as err:
            raise RuntimeError("Failed to save running config to disk") from err

    # pylint: disable-next=unused-argument
    def sync(self, *args: Any, **kwargs: Any) -> None:
        """
        Sync BIG-IP node
        """
        try:
            cmd: str = f"run /cm config-sync to-group {self.sync_group}"
            logging.debug("`%s`", cmd)
            self.conn.run(cmd)
        except UnexpectedExit as err:
            raise RuntimeError(
                f"Failed to sync to device group `{self.sync_group}`"
            ) from err

    @staticmethod
    def entrypoint(
        *, args: argparse.Namespace, certificate_bundle: CertificateBundle
    ) -> None:
        """
        Run tasks for this deployer
        """
        bigip_certificate_bundle: BigipCertificateBundle = BigipCertificateBundle(
            path=args.renewed_lineage, name=args.cert_name
        )
        profile: Optional[CertProfile]
        if args.profile_name is not None and args.profile_type is not None:
            profile = CertProfile(name=args.profile_name, type=args.profile_type)
        else:
            profile = None

        deployer: BigipDeployer = BigipDeployer(
            certificate_bundle=bigip_certificate_bundle,
            dest_temp_dir=args.dest_temp_dir,
            host=args.host,
            profile=profile,
            sync_group=args.sync_group,
        )
        workflow = deployer.get_workflow()
        if args.dry_run:
            print("# Running in dry run mode. Will not run actual deployment tasks.")
        for task in workflow:
            if args.dry_run:
                print(f"Would run task: {task.name}")
            else:
                logging.info("Running workflow step: %s", task.name)
                task.execute()



"""
main

manipulates argv in order to pass the run on to the framework as
`certbot-deployer <subcommand>
"""

import sys

from certbot_deployer import main as framework_main
from ..certbot_deployer_bigip import BigipDeployer


def main(
    argv: list = sys.argv[1:],
) -> None:
    # pylint: disable=line-too-long
    """
    main
    """
    new_argv = [BigipDeployer.subcommand] + argv
    framework_main(deployers=[BigipDeployer], argv=new_argv)



"""
Unit tests for BigipDeployer
"""

import argparse
import os
import posixpath

from pathlib import Path
from typing import Any, List, Optional, Tuple, Type

import pytest
import scp as paramiko_scp  # type: ignore

from invoke.exceptions import UnexpectedExit
from invoke.runners import Result

from certbot_deployer import (
    CERT,
    CERT_FILENAME,
    FULLCHAIN,
    FULLCHAIN_FILENAME,
    INTERMEDIATES_FILENAME,
    KEY_FILENAME,
)
from certbot_deployer import CertificateComponent, Deployer
from certbot_deployer.test_helpers import generate_self_signed_cert
import certbot_deployer_bigip._main as plugin_main
from certbot_deployer_bigip.certbot_deployer_bigip import (
    BigipCertificateBundle,
    BigipDeployer,
    BigipTask,
    CertProfile,
)


HOST: str = "something.domain.tld"
KEY: str = "key"
TEST_PROFILE: CertProfile = CertProfile(name="test_profile", type="client-ssl")
TEST_SYNC_GROUP: str = "test_sync_group"


@pytest.fixture(name="bigip_certificate_bundle", scope="function")
def fixture_bigip_certificate_bundle(tmp_path: Path) -> BigipCertificateBundle:
    """
    Create a temporary Certbot "certificate bundle" in a temp dir for testing.

    It generates the following files:
        - CERT_FILENAME
        - FULLCHAIN_FILENAME
        - INTERMEDIATES_FILENAME
        - KEY_FILENAME

    Note:
        This fixture creates a self-signed certificate. It need only be
        parseable - in production, its cryptographic validity should have just
        been verified by Certbot itself before this script is called.

        The other files are plaintext.

    Returns:
        BigipCertificateBundle: An instance wrapping the temporary certificate
        bundle directory.
    """
    with open(tmp_path / CERT_FILENAME, "w", encoding="utf-8") as certfile:
        certfile.write(generate_self_signed_cert())
    with open(tmp_path / FULLCHAIN_FILENAME, "w", encoding="utf-8") as certfile:
        certfile.write("fullchain")
    with open(tmp_path / INTERMEDIATES_FILENAME, "w", encoding="utf-8") as certfile:
        certfile.write("intermediates")
    with open(tmp_path / KEY_FILENAME, "w", encoding="utf-8") as certfile:
        certfile.write("key")
    return BigipCertificateBundle(path=str(tmp_path))


@pytest.fixture(name="bigip_deployer_base", scope="function")
def fixture_bigip_deployer(
    bigip_certificate_bundle: BigipCertificateBundle,
) -> BigipDeployer:
    """
    Return a BigipDeployer for testing.

    This fixture initializes a BigipDeployer with a temporary certificate bundle,
    but without a sync group or profile for simplicity.
    """
    deployer: BigipDeployer = BigipDeployer(
        host=HOST,
        dest_temp_dir="/remote/path/",
        certificate_bundle=bigip_certificate_bundle,
        sync_group=None,
        profile=None,
    )
    return deployer


@pytest.fixture(name="bigip_deployer_with_profile", scope="function")
def fixture_bigip_deployer_with_profile(
    bigip_certificate_bundle: BigipCertificateBundle,
) -> BigipDeployer:
    """
    Return a BigipDeployer for testing with a specified certificate profile.

    A sample CertProfile object (`TEST_PROFILE`) is assigned to the deployer.
    """
    deployer: BigipDeployer = BigipDeployer(
        host=HOST,
        dest_temp_dir="/remote/path/",
        certificate_bundle=bigip_certificate_bundle,
        sync_group=None,
        profile=TEST_PROFILE,
    )
    return deployer


@pytest.fixture(name="bigip_deployer_with_sync_group", scope="function")
def fixture_bigip_deployer_with_sync_group(
    bigip_certificate_bundle: BigipCertificateBundle,
) -> BigipDeployer:
    """
    Return a BigipDeployer for testing with a specified sync group.

    A sample sync group (`TEST_SYNC_GROUP`) is assigned to the deployer.
    """
    deployer: BigipDeployer = BigipDeployer(
        host=HOST,
        dest_temp_dir="/remote/path/",
        certificate_bundle=bigip_certificate_bundle,
        sync_group=TEST_SYNC_GROUP,
        profile=None,
    )
    return deployer


@pytest.fixture(name="deployer_base_with_expected_tasks", scope="function")
def fixture_deployer_base_with_expected_tasks(
    bigip_deployer_base: BigipDeployer,
) -> Tuple[BigipDeployer, List[BigipTask]]:
    """
    Create a basic BigipDeployer instance with its associated workflow tasks.

    This fixture initializes a BigipDeployer configured with a temporary certificate bundle.
    It also constructs a list of expected BigipTask objects representing the
    deployment steps.

    Returns:
        Tuple[BigipDeployer, List[BigipTask]]: A tuple containing the
        BigipDeployer instance and the list of expected deployment tasks.
    """
    deployer = bigip_deployer_base
    tasks: List[BigipTask] = [
        BigipTask(
            exec_function=deployer.put_bigip_file,
            exec_kwargs={"component": deployer.certificate_bundle.fullchain},
        ),
        BigipTask(
            exec_function=deployer.install_cert,
            exec_kwargs={"component": deployer.certificate_bundle.fullchain},
        ),
        BigipTask(
            exec_function=deployer.verify_cert_installed,
            exec_kwargs={"component": deployer.certificate_bundle.fullchain},
        ),
        BigipTask(
            exec_function=deployer.put_bigip_file,
            exec_kwargs={"component": deployer.certificate_bundle.key},
        ),
        BigipTask(
            exec_function=deployer.install_cert,
            exec_kwargs={"component": deployer.certificate_bundle.key},
        ),
        BigipTask(
            exec_function=deployer.verify_cert_installed,
            exec_kwargs={"component": deployer.certificate_bundle.key},
        ),
        BigipTask(
            exec_function=deployer.zero_bigip_file,
            exec_kwargs={"component": deployer.certificate_bundle.fullchain},
        ),
        BigipTask(
            exec_function=deployer.zero_bigip_file,
            exec_kwargs={"component": deployer.certificate_bundle.key},
        ),
        BigipTask(
            exec_function=deployer.save,
        ),
    ]
    return (deployer, tasks)


@pytest.fixture(name="deployer_with_profile_and_expected_tasks", scope="function")
def fixture_deployer_with_profile_and_expected_tasks(
    deployer_base_with_expected_tasks: Tuple[BigipDeployer, List[BigipTask]],
) -> Tuple[BigipDeployer, List[BigipTask]]:
    """
    Extend the base BigipDeployer fixture with certificate profile configuration.

    This fixture builds upon the basic deployer by assigning a certification
    profile and appending the profile management task to the expected workflow.

    Returns:
        Tuple[BigipDeployer, List[BigipTask]]: A tuple containing the
        BigipDeployer instance with an assigned certificate profile and the
        updated task list.
    """
    deployer: BigipDeployer
    tasks: List[BigipTask]
    deployer, tasks = deployer_base_with_expected_tasks
    tasks.append(
        BigipTask(
            exec_function=deployer.manage_profile,
            exec_kwargs={"component": deployer.certificate_bundle.fullchain},
        )
    )
    deployer.profile = TEST_PROFILE
    return (deployer, tasks)


@pytest.fixture(name="deployer_with_sync_group_and_expected_tasks", scope="function")
def fixture_deployer_with_sync_group_and_expected_tasks(
    deployer_base_with_expected_tasks: Tuple[BigipDeployer, List[BigipTask]],
) -> Tuple[BigipDeployer, List[BigipTask]]:
    """
    Extend the base BigipDeployer with synchronization group configuration.

    This fixture sets a sync group for the deployer and adds sync-related tasks
    to the expected workflow.

    Returns:
        Tuple[BigipDeployer, List[BigipTask]]: A tuple containing the
        BigipDeployer instance with a configured sync group and the updated
        task list.
    """
    deployer: BigipDeployer
    tasks: List[BigipTask]
    deployer, tasks = deployer_base_with_expected_tasks
    tasks.insert(
        0,
        BigipTask(exec_function=deployer.verify_sync),
    )
    tasks.extend(
        [
            BigipTask(
                exec_function=deployer.sync,
            ),
        ]
    )
    deployer.sync_group = "mysyncgroup"
    return (deployer, tasks)


@pytest.fixture(
    name="deployer_with_sync_group_and_profile_and_expected_tasks", scope="function"
)
def fixture_deployer_with_sync_group_and_profile_and_expected_tasks(
    deployer_with_sync_group_and_expected_tasks: Tuple[BigipDeployer, List[BigipTask]],
) -> Tuple[BigipDeployer, List[BigipTask]]:
    """
    Extend the base BigipDeployer with synchronization group configuration AND
    profile configuration.

    This fixture sets a sync group and profile for the deployer and appends the
    sync task to the expected workflow.

    Returns:
        Tuple[BigipDeployer, List[BigipTask]]: A tuple containing the
        BigipDeployer instance with a configured sync group and profile and the updated
        task list.
    """
    deployer: BigipDeployer
    tasks: List[BigipTask]
    deployer, tasks = deployer_with_sync_group_and_expected_tasks
    deployer.profile = TEST_PROFILE
    tasks.insert(
        len(tasks) - 1,
        BigipTask(
            exec_function=deployer.manage_profile,
            exec_kwargs={"component": deployer.certificate_bundle.fullchain},
        ),
    )
    return (deployer, tasks)


@pytest.fixture(name="deployer_and_expected_tasks", scope="function")
def fixture_deployer_and_expected_tasks(
    request: pytest.FixtureRequest,
) -> Tuple[BigipDeployer, List[BigipTask]]:
    """
    Dynamically load and return a BigipDeployer along with its expected workflow tasks.

    This fixture factory utilizes the parameterized fixture name passed by
    indirect pytest parametrization and returns the corresponding deployer
    instance and expected tasks.

    Returns:
        Tuple[BigipDeployer, List[BigipTask]]: The deployer configuration and
        its expected task list.
    """
    return request.getfixturevalue(request.param)


def test_subcommand() -> None:
    """
    Test that the BigipDeployer's subcommand class variable is correctly set.
    """
    assert BigipDeployer.subcommand == "bigip"


def test_static_argparse_post() -> None:
    """
    Test the argument validation logic in BigipDeployer.argparse_post().

    Verifies that:
        - An argparse.ArgumentTypeError is raised if exactly one of the
          following arguments is provided: profile or profile_type.
        - No exception is raised if both are provided or if neither is provided.
    """
    args: argparse.Namespace

    # Verify `--host` is "required" by validation in post
    args = argparse.Namespace()
    with pytest.raises(argparse.ArgumentTypeError):
        BigipDeployer.argparse_post(args=args)

    # Raise when only one of `profile_name` and `profile_type` given
    args = argparse.Namespace(host="val", profile_name=None, profile_type="val")
    with pytest.raises(argparse.ArgumentTypeError):
        BigipDeployer.argparse_post(args=args)
    args = argparse.Namespace(host="val", profile_name="val", profile_type=None)
    with pytest.raises(argparse.ArgumentTypeError):
        BigipDeployer.argparse_post(args=args)

    # Succeed otherwise
    args = argparse.Namespace(host="val", profile_name="val", profile_type="val")
    BigipDeployer.argparse_post(args=args)
    args = argparse.Namespace(host="val", profile_name=None, profile_type=None)
    BigipDeployer.argparse_post(args=args)


@pytest.mark.parametrize(
    # Parametrize fixtures that provide a BigipDeployer and a list of expected
    # workflow Tasks
    "deployer_and_expected_tasks",
    [
        "deployer_base_with_expected_tasks",
        "deployer_with_profile_and_expected_tasks",
        "deployer_with_sync_group_and_expected_tasks",
        "deployer_with_sync_group_and_profile_and_expected_tasks",
    ],
    indirect=True,
    ids=[
        "deployer_base_with_expected_tasks",
        "deployer_with_profile_and_expected_tasks",
        "deployer_with_sync_group_and_expected_tasks",
        "deployer_with_sync_group_and_profile_and_expected_tasks",
    ],
)
def test_workflow(
    deployer_and_expected_tasks: Tuple[BigipDeployer, List[BigipTask]],
) -> None:
    """
    Test that the deployer workflow matches the expected sequence of tasks.

    This test retrieves the workflow generated by BigipDeployer.get_workflow()
    and verifies it matches the expected task list.
    """
    deployer: BigipDeployer
    tasks: List[BigipTask]
    deployer, expected_tasks = deployer_and_expected_tasks
    tasks = deployer.get_workflow()
    assert len(tasks) == len(expected_tasks)
    for task, expected_task in zip(tasks, expected_tasks):
        assert task == expected_task


def test_verify_sync(
    monkeypatch: pytest.MonkeyPatch,
    bigip_deployer_base: BigipDeployer,
) -> None:
    """
    Test the verify_sync function by simulating BIG-IP sync status via tmsh commands.
    """

    bigip_deployer = bigip_deployer_base

    # pylint: disable-next=too-few-public-methods
    class DummyResponse:
        """dummy"""

        def __init__(self, stdout: str) -> None:
            self.stdout: str = stdout

    monkeypatch.setattr(
        bigip_deployer.conn, "run", lambda cmd: DummyResponse("Status: In Sync")
    )
    # Expect no exception when sync status is "In Sync".
    bigip_deployer.verify_sync()

    monkeypatch.setattr(
        bigip_deployer.conn,
        "run",
        lambda cmd: DummyResponse("Status: Bad and not good"),
    )
    with pytest.raises(RuntimeError):
        bigip_deployer.verify_sync()


def test_put_file(
    monkeypatch: pytest.MonkeyPatch,
    bigip_deployer_base: BigipDeployer,
) -> None:
    """
    Test deploying files to the remote BIG-IP via scp.
    """
    bigip_deployer = bigip_deployer_base
    calls: List[Tuple[str, str]] = []

    def fake_scp(localpath: str, remotepath: str) -> None:
        nonlocal calls
        calls.append((localpath, remotepath))

    monkeypatch.setattr(bigip_deployer, "_scp", fake_scp)
    component: CertificateComponent = bigip_deployer.certificate_bundle.cert
    bigip_deployer.put_bigip_file(component=component)
    assert len(calls) == 1
    assert calls[0] == (
        component.path,
        posixpath.join(bigip_deployer.dest_temp_dir, component.filename),
    )


@pytest.mark.parametrize(
    "transport_ready",
    [
        True,
        False,
    ],
    ids=[
        "Transport not created yet before `_scp` is called",
        "Transport already created before `_scp` is called",
    ],
)
def test_scp(
    monkeypatch: pytest.MonkeyPatch,
    bigip_deployer_base: BigipDeployer,
    transport_ready: bool,
) -> None:
    """
    Mock the paramiko scp and ensure our scp method tries to
    upload the right files to the right host

    additionally verify that the connection transport is opened when transport
    is not ready
    """
    bigip_deployer = bigip_deployer_base
    localfile: str = "LOCALFILE"
    remotefile: str = "REMOTEFILE"
    put_count: int = 0
    transport_open_count: int = 0

    monkeypatch.setattr(bigip_deployer.conn, "transport", transport_ready)

    def mock_conn_open() -> None:
        nonlocal transport_open_count
        transport_open_count += 1

    # pylint: disable=missing-function-docstring
    # pylint: disable=missing-class-docstring
    # pylint: disable=too-few-public-methods
    class MockSCPClient:
        def __init__(self, transport: str) -> None:
            pass

        # pylint: disable-next=unused-argument
        def put(self, localpath: str, remotepath: str) -> None:
            nonlocal put_count
            put_count += 1
            assert localpath == localfile
            assert remotepath == remotefile

        def close(self) -> None:
            pass

    # pylint: enable=too-few-public-methods
    # pylint: enable=missing-class-docstring
    # pylint: enable=missing-function-docstring

    monkeypatch.setattr(paramiko_scp, "SCPClient", MockSCPClient)
    monkeypatch.setattr(bigip_deployer.conn, "open", mock_conn_open)
    # pylint: disable-next=protected-access
    bigip_deployer._scp(localfile, remotefile)
    assert put_count == 1
    if not transport_ready:
        assert (
            transport_open_count == 1
        ), "Connection was not yet opened and should have been opened before doing SCP"
    else:
        assert (
            transport_open_count == 0
        ), "Connection was already opened and should not have been opened before doing SCP"


@pytest.mark.parametrize(
    # Parametrize to test the different types of certificate file that we try
    # to install on the BIG-IP
    "component_label, bigip_cert_type",
    [
        (CERT, CERT),
        (FULLCHAIN, CERT),
        (KEY, KEY),
    ],
    ids=[
        "Install CERT as a certificate",
        "Install FULLCHAIN as a certificate",
        "Install KEY as a key",
    ],
)
def test_install_cert(
    monkeypatch: pytest.MonkeyPatch,
    bigip_deployer_base: BigipDeployer,
    component_label: str,
    bigip_cert_type: str,
) -> None:
    """
    Test installing certificates on the remote BIG-IP via tmsh commands.
    """
    bigip_deployer = bigip_deployer_base
    commands: List[str] = []

    def fake_run(cmd: str) -> None:
        nonlocal commands
        commands.append(cmd)

    monkeypatch.setattr(bigip_deployer.conn, "run", fake_run)
    component: CertificateComponent = getattr(
        bigip_deployer.certificate_bundle, component_label
    )
    remote_path: str = posixpath.join(bigip_deployer.dest_temp_dir, component.filename)
    expected_cmd: str = (
        f"install /sys crypto {bigip_cert_type} {bigip_deployer.certificate_bundle.name} "
        f"from-local-file {remote_path}"
    )
    bigip_deployer.install_cert(component=component)
    assert len(commands) == 1
    assert commands[0] == expected_cmd

    def fake_run_fail(cmd: str) -> None:
        raise UnexpectedExit(Result())

    monkeypatch.setattr(bigip_deployer.conn, "run", fake_run_fail)
    with pytest.raises(RuntimeError):
        bigip_deployer.install_cert(component=component)


@pytest.mark.parametrize(
    # Parametrize to test the different types of certificate file that we try
    # to install on the BIG-IP
    "component_label, bigip_cert_type",
    [
        (CERT, CERT),
        (FULLCHAIN, CERT),
        (KEY, KEY),
    ],
    ids=[
        "Verify CERT as a certificate",
        "Verify FULLCHAIN as a certificate",
        "Verify KEY as a key",
    ],
)
def test_verify_cert_installed(
    monkeypatch: pytest.MonkeyPatch,
    bigip_deployer_base: BigipDeployer,
    component_label: str,
    bigip_cert_type: str,
) -> None:
    """
    Test verifying certs on the BIG-IP
    """
    bigip_deployer = bigip_deployer_base
    commands: List[str] = []

    component: CertificateComponent = getattr(
        bigip_deployer.certificate_bundle, component_label
    )
    expected_cmd: str = (
        f"list /sys crypto {bigip_cert_type} {bigip_deployer.certificate_bundle.name}"
    )

    # pylint: disable-next=too-few-public-methods
    class FakeResult:
        """Fake the results from invoke's run()"""

        def __init__(self, stdout: str) -> None:
            self.stdout: str = stdout

    def fake_run(cmd: str) -> FakeResult:
        nonlocal commands
        commands.append(cmd)
        return FakeResult(bigip_deployer.certificate_bundle.fingerprint)

    monkeypatch.setattr(bigip_deployer.conn, "run", fake_run)
    bigip_deployer.verify_cert_installed(component=component)
    assert len(commands) == 1
    assert commands[0] == expected_cmd

    def fake_run_no_fingerprint_in_output(_: str) -> FakeResult:
        return FakeResult("no fingerprints to be found here")

    monkeypatch.setattr(bigip_deployer.conn, "run", fake_run_no_fingerprint_in_output)
    if bigip_cert_type == CERT:
        with pytest.raises(RuntimeError):
            bigip_deployer.verify_cert_installed(component=component)

    def fake_run_unexpected_exit(cmd: str) -> None:
        raise UnexpectedExit(Result())

    monkeypatch.setattr(bigip_deployer.conn, "run", fake_run_unexpected_exit)
    with pytest.raises(RuntimeError):
        bigip_deployer.verify_cert_installed(component=component)


def test_zero_file(
    monkeypatch: pytest.MonkeyPatch,
    bigip_deployer_base: BigipDeployer,
) -> None:
    """
    Test deploying empty files to the remote BIG-IP via scp.
    """
    bigip_deployer = bigip_deployer_base
    calls: List[Tuple[str, str]] = []

    def fake_scp(localpath: str, remotepath: str) -> None:
        assert os.path.getsize(localpath) == 0
        nonlocal calls
        calls.append((localpath, remotepath))

    monkeypatch.setattr(bigip_deployer, "_scp", fake_scp)

    component: CertificateComponent = bigip_deployer.certificate_bundle.cert
    bigip_deployer.zero_bigip_file(component=component)
    assert len(calls) == 1
    call: Tuple[str, str] = calls[0]
    called_remote_filepath: str
    _, called_remote_filepath = call
    assert called_remote_filepath == posixpath.join(
        bigip_deployer.dest_temp_dir, component.filename
    )


@pytest.mark.parametrize(
    # The method being tested will `run()` to check if the profile exists, and
    # *then* run either `create` or `modify`
    # Parametrize to tell the test which steps to raise on, and what operations
    # to expect
    "runs_to_raise_on, expected_operations, expect_exception",
    [
        # Successes
        ([], ["list", "modify"], False),
        ([0], ["list", "create"], False),
        # Failures
        ([1], ["list", "modify"], True),
        ([0, 1], ["list", "create"], True),
    ],
    ids=[
        "`list` to see if profile exists, and then succeed at `modify`",
        "`list` to see if profile exists (nope), and then succeed at `create`",
        #
        "`list` to see if profile exists, and then raise when attempting `modify`",
        "`list` to see if profile exists, and then raise when attempting `create`",
    ],
)
def test_manage_profile(
    monkeypatch: pytest.MonkeyPatch,
    bigip_deployer_with_profile: BigipDeployer,
    runs_to_raise_on: List[int],
    expected_operations: List[str],
    expect_exception: bool,
) -> None:
    """
    Test managing (creating or modifying) profiles on BIG-IP.

    This test verifies the behavior of the `manage_profile` method, which
    checks if a profile exists on the BIG-IP device and either modifies it
    or creates it if it does not exist.

    Parameters:
        runs_to_raise_on (List[int]): Indices of commands on which to simulate failure
        expected_operations (List[str]): Expected sequence of operations
        expect_exception (bool): Whether an exception is expected on `run()`
    """
    bigip_deployer = bigip_deployer_with_profile
    commands: List[str] = []

    def fake_run(cmd: str) -> None:
        nonlocal commands
        commands.append(cmd)
        run_count: int = len(commands) - 1
        if run_count in runs_to_raise_on:
            raise UnexpectedExit(Result())

    monkeypatch.setattr(bigip_deployer.conn, "run", fake_run)
    if expect_exception:
        with pytest.raises(RuntimeError):
            bigip_deployer.manage_profile()
    else:
        bigip_deployer.manage_profile()

    assert len(expected_operations) == len(commands)

    for command, expected_operation in zip(commands, expected_operations):
        operation: str
        args: str
        operation, args = command.split(" ", 1)
        assert operation == expected_operation
        assert bigip_deployer.profile is not None
        if operation == "list":
            assert args == (
                f"/ltm profile {bigip_deployer.profile.type} "
                f"{bigip_deployer.profile.name}"
            )
        else:
            assert args == (
                f"/ltm profile {bigip_deployer.profile.type} "
                f"{bigip_deployer.profile.name} cert "
                f"{bigip_deployer.certificate_bundle.name} key "
                f"{bigip_deployer.certificate_bundle.name}"
            )


def test_sync(
    monkeypatch: pytest.MonkeyPatch,
    bigip_deployer_with_sync_group: BigipDeployer,
) -> None:
    """
    Test the `sync` method of BigipDeployer.

    This test verifies that the `sync` method correctly runs the appropriate
    command to synchronize the BIG-IP configuration and handles errors
    appropriately.
    """
    commands: List[str] = []
    bigip_deployer = bigip_deployer_with_sync_group

    def fake_run(cmd: str) -> None:
        nonlocal commands
        commands.append(cmd)

    monkeypatch.setattr(bigip_deployer.conn, "run", fake_run)
    expected_cmd: str = f"run /cm config-sync to-group {bigip_deployer.sync_group}"

    # Test for a successful run
    bigip_deployer.sync()
    assert len(commands) == 1
    assert commands[0] == expected_cmd

    commands = []

    def fake_run_fail(cmd: str) -> None:
        raise UnexpectedExit(Result())

    monkeypatch.setattr(bigip_deployer.conn, "run", fake_run_fail)
    with pytest.raises(RuntimeError):
        bigip_deployer.sync()


def test_static_entrypoint(
    monkeypatch: pytest.MonkeyPatch,
    bigip_certificate_bundle: BigipCertificateBundle,
) -> None:
    """
    Test the `entrypoint` method of BigipDeployer.

    This test ensures that the `entrypoint` method correctly executes the
    workflow tasks defined for the deployer.
    """
    run_count: int = 0

    def fake_task() -> None:
        nonlocal run_count
        run_count += 1

    tasks: List[BigipTask] = [
        BigipTask(exec_function=fake_task),
        BigipTask(exec_function=fake_task),
    ]

    def fake_get_workflow(_: Any) -> List[BigipTask]:
        return tasks

    monkeypatch.setattr(BigipDeployer, "get_workflow", fake_get_workflow)
    args: argparse.Namespace = argparse.Namespace(
        cert_name=None,
        dest_temp_dir=None,
        dry_run=False,
        host="somewhere.domain.tld",
        profile_name=None,
        profile_type=None,
        renewed_lineage=bigip_certificate_bundle.path,
        sync_group=None,
        user=None,
    )
    BigipDeployer.entrypoint(args=args, certificate_bundle=bigip_certificate_bundle)
    assert run_count == len(tasks)


def test_dry_run(
    monkeypatch: pytest.MonkeyPatch,
    capsys: Any,
    bigip_certificate_bundle: BigipCertificateBundle,
) -> None:
    """
    Verify that our entrypoint prints out what it *would* do and does not run anything
    """
    run_count: int = 0

    def fake_task() -> None:
        nonlocal run_count
        run_count += 1

    task_names: List[str] = [
        "mytask1",
        "mytask2",
    ]
    tasks: List[BigipTask] = []
    for task_name in task_names:
        tasks.append(BigipTask(exec_function=fake_task, name=task_name))

    def fake_get_workflow(_: Any) -> List[BigipTask]:
        return tasks

    monkeypatch.setattr(BigipDeployer, "get_workflow", fake_get_workflow)
    args: argparse.Namespace = argparse.Namespace(
        cert_name=None,
        dest_temp_dir=None,
        dry_run=True,
        host="somewhere.domain.tld",
        profile_name=None,
        profile_type=None,
        renewed_lineage=bigip_certificate_bundle.path,
        sync_group=None,
        user=None,
    )
    BigipDeployer.entrypoint(args=args, certificate_bundle=bigip_certificate_bundle)

    assert run_count == 0
    stdout: str = capsys.readouterr().out
    for task_name in task_names:
        assert task_name in stdout
    print(stdout)


def test_main_delegation(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Verify that our `main()` hands control off to the framework when called
    directly by mocking the framework's `main()` and comparing the passed
    args/deployers
    """
    called_argv: list = []
    called_deployers: Optional[List[Type[Deployer]]]

    def fake_framework_main(
        argv: list, deployers: Optional[List[Type[Deployer]]] = None
    ) -> None:
        nonlocal called_argv
        nonlocal called_deployers
        called_argv = argv
        called_deployers = deployers

    argv: List[str] = ["-h"]
    expected_argv: List[str] = [BigipDeployer.subcommand, "-h"]
    expected_deployers: Optional[List[Type[Deployer]]] = [BigipDeployer]

    monkeypatch.setattr(
        plugin_main,
        "framework_main",
        fake_framework_main,
    )
    plugin_main.main(argv=argv)
    assert called_argv == expected_argv
    assert called_deployers == expected_deployers
